--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  FILE         | parameterzied_priority_encoder.vhd   --
--  -------------+------------------------------------  --
--  DESCRIPTION  | Parameterized priority encoder       --
--  -------------+------------------------------------  --
--  AUTHOR       | Pavle Belanovic                      --
--  -------------+------------------------------------  --
--  DATE         | 20 June 2002                         --
--                                                      --
--======================================================--

--******************************************************************************--
--                                                                              --
--	Copyright (C) 2000 Pavle Belanovic                                          --
--                                                                              --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                              --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                              --
--	You should have received a copy of the GNU General Public License           --
--	along with this program; if not, write to the Free Software                 --
--	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. --
--                                                                              --
--******************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

-- float
library fp_lib;
use fp_lib.float_pkg.all;

----------------------------------------------------------
--           Parameterized priority encoder             --
----------------------------------------------------------
entity parameterized_priority_encoder is
	generic
	(
		man_bits		:	integer	:=	0;
		shift_bits	:	integer	:=	0
	);
	port
	(
		--inputs
		MAN_IN				:	in		std_logic_vector(man_bits-1 downto 0);
		--outputs
		SHIFT					:	out		std_logic_vector(shift_bits-1 downto 0)	:=	(others=>'0');
		EXCEPTION_OUT	:	out	std_logic																	:=	'0'
	);
end parameterized_priority_encoder;

----------------------------------------------------------
--           Parameterized priority encoder             --
----------------------------------------------------------
architecture parameterized_priority_encoder_arch of parameterized_priority_encoder is
	--FUNCTIONS
	--function to calculate ceil(log2(n))
	function ceil_log2 (input: integer) return integer is
		variable max		:	integer :=1; -- exp=0 => max=2^0=1
		variable exp		:	integer :=0;
	begin
		log_loop: while max < input loop
			exp	:=	exp+1;	-- increment exponent
			max	:=	max*2;	-- max = 2^exp
		end loop log_loop;
		return exp; -- exponent that gives (2^exp) >= input
	end ceil_log2;
	--TYPES
	type or_pipe is
		array (0 to shift_bits-1) of std_logic_vector((2**(shift_bits-1))-1 downto 0);
	--SIGNALS
	signal	all_or				:	or_pipe;
	signal	im						:	std_logic_vector(man_bits-1 downto 0)	:=	(others=>'0');
	signal	nim						:	std_logic_vector(man_bits-1 downto 0)	:=	(others=>'0');
	signal	shift_one_hot	:	std_logic_vector((2**shift_bits)-1 downto 0)	:=	(others=>'0');
	
begin
	--check for defective module
	defective:	if(shift_bits/=ceil_log2(man_bits)) generate
		SHIFT	<=	(others=>'0');
	end generate;--defective
	--correctly instantiated
	correct	:	if(shift_bits=ceil_log2(man_bits))	generate
		--=========== MAN_IN ---> SHIFT_ONE_HOT =============
		im(man_bits-1) <=	MAN_IN(man_bits-1);--rename
		--array of or gates
		or_gates	:	for i in man_bits-2 downto 0 generate
			im(i)	<=	im(i+1) OR	MAN_IN(i);
		end generate;--or gates
		--array of inverters
		nim	<=	NOT(im);
		--array of and gates
		and_gates	:	for j in man_bits-1 downto 1 generate
			shift_one_hot(man_bits-j)	<=	nim(j)	AND im(j-1);
		end generate;--and gates
		--ends of shift_one_hot
		shift_one_hot(0)					<=	im(man_bits-1);
		--fill up shift_one_hot with zeros
		fill_shift_one_hot_with_zeros	:	if(man_bits/=2**shift_bits) generate
			shift_one_hot((2**shift_bits)-1 downto man_bits)	<=	(others=>'0');
		end generate;--fill shift_one_hot_with zeros
		--exception_out, all-zero mantissa
		EXCEPTION_OUT	<=	nim(0);
		--=========== SHIFT_ONE_HOT ---> SHIFT ==============
		--several multiple-input OR gates
		wide_or_gates	:	for k in 0 to shift_bits-1 generate
			--wire all_or
			wire_all_or	:	for b in 0 to (2**(shift_bits-k-1))-1 generate
				all_or(k)(((b+1)*(2**k))-1 downto (b*(2**k)))	<=	shift_one_hot((2**(k+1))*(b+1)-1 downto (2**k)*(2*b+1));
			end generate;--wire all_or
			--instantiate the gate
			wide_or_gate	:	parameterized_or_gate
				generic map
				(
					bits	=>	2**(shift_bits-1)
				)
				port map
				(
					A			=>	all_or(k),
					O			=>	SHIFT(k)
				);
		end generate;--wide or gates
	end generate;--correctly instantiated
end parameterized_priority_encoder_arch; -- end of architecture
