--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  FILE         | fp_add.vhd                           --
--  -------------+------------------------------------  --
--  DESCRIPTION  | Floating-point addition module       --
--  -------------+------------------------------------  --
--  AUTHOR       | Pavle Belanovic                      --
--  -------------+------------------------------------  --
--  DATE         | 20 June 2002                         --
--                                                      --
--======================================================--

--******************************************************************************--
--                                                                              --
--	Copyright (C) 2000 Pavle Belanovic                                          --
--                                                                              --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                              --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                              --
--	You should have received a copy of the GNU General Public License           --
--	along with this program; if not, write to the Free Software                 --
--	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. --
--                                                                              --
--******************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
--use IEEE.std_logic_signed.all;

-- float
library work;
use work.float_pkg.all;

----------------------------------------------------------
--               Floating Point Adder                   --
----------------------------------------------------------
entity fp_add is
	generic
	(
		exp_bits			:	integer	:=	8;
		man_bits			:	integer	:=	23
	);
	port
	(
		--inputs
		OP1						:	in		std_logic_vector(man_bits+exp_bits downto 0);
		OP2						:	in		std_logic_vector(man_bits+exp_bits downto 0);
		READY					:	in		std_logic;
		EXCEPTION_IN	:	in		std_logic;
		CLK						:	in		std_logic;
		--outputs
		RESULT				:	out		std_logic_vector(man_bits+exp_bits+1 downto 0)	:=	(others=>'0');
		EXCEPTION_OUT	:	out		std_logic																				:=	'0';
		DONE					:	out		std_logic																				:=	'0'
	);
end fp_add;

----------------------------------------------------------
--               Floating Point Adder                   --
----------------------------------------------------------
architecture fp_add_arch of fp_add is
	signal	rdy1				:	std_logic																			:=	'0';
	signal	rdy2				:	std_logic																			:=	'0';
	signal	rdy3				:	std_logic																			:=	'0';
	signal	exc1				:	std_logic																			:=	'0';
	signal	exc2				:	std_logic																			:=	'0';
	signal	exc3				:	std_logic																			:=	'0';
	signal	s1					:	std_logic																			:=	'0';
	signal	s2					:	std_logic																			:=	'0';
	signal	op					:	std_logic																			:=	'0';
	signal	op_2				:	std_logic																			:=	'0';
	signal	s1_2				:	std_logic																			:=	'0';
	signal	s1_3				:	std_logic																			:=	'0';
	signal	gt					:	std_logic																			:=	'0';
	signal	lt					:	std_logic																			:=	'0';
	signal	eq					:	std_logic																			:=	'0';
	signal	eq1					:	std_logic																			:=	'0';
	signal	eq2					:	std_logic																			:=	'0';
	signal	eq3					:	std_logic																			:=	'0';
	signal	enable			:	std_logic																			:=	'0';
	signal	e1					:	std_logic_vector(exp_bits-1 downto 0)					:=	(others=>'0');
	signal	e2					:	std_logic_vector(exp_bits-1 downto 0)					:=	(others=>'0');
	signal	e1_2				:	std_logic_vector(exp_bits-1 downto 0)					:=	(others=>'0');
	signal	e1_3				:	std_logic_vector(exp_bits-1 downto 0)					:=	(others=>'0');
	signal	f2					:	std_logic_vector(man_bits-1 downto 0)					:=	(others=>'0');
	signal	f1					:	std_logic_vector(man_bits downto 0)						:=	(others=>'0');
	signal	f1_2				:	std_logic_vector(man_bits downto 0)						:=	(others=>'0');
	signal	f2_2				:	std_logic_vector(man_bits downto 0)						:=	(others=>'0');
	signal	f_out				:	std_logic_vector(man_bits downto 0)						:=	(others=>'0');
	signal	large				:	std_logic_vector(man_bits+exp_bits downto 0)	:=	(others=>'0');
	signal	small				:	std_logic_vector(man_bits+exp_bits downto 0)	:=	(others=>'0');
	signal	n_op2				:	std_logic_vector(man_bits+exp_bits downto 0)	:=	(others=>'0');
begin
	--instances of all the components
	equal: parameterized_comparator
		generic map
		(
			bits			=>	exp_bits+man_bits+1
		)
		port map
		(
			A					=>	OP1,
			B					=>	n_op2,
			A_GT_B		=>	gt,
			A_EQ_B		=>	eq,
			A_LT_B		=>	lt
		);
	sw: swap
		generic map
		(
			exp_bits			=>	exp_bits,
			man_bits			=>	man_bits
		)
		port map
		(
			--inputs
			IN1						=>	OP1,
			IN2						=>	OP2,
			READY					=>	READY,
			CLK						=>	CLK,
			--outputs
			OUT1					=>	large,
			OUT2					=>	small,
			DONE					=>	rdy1
		);
	 
	s_a: shift_adjust 
		generic map
		(
			exp_bits			=>	exp_bits,
			man_bits			=>	man_bits
		)
		port map
		(
			--inputs
			E1						=>	e1,
			E2						=>	e2,
			F2						=>	f2,
			READY					=>	rdy1,
			CLK						=>	CLK,
			--outputs
			E_OUT					=>	e1_2,
			F_OUT					=>	f2_2,
			DONE					=>	rdy2
		);
	 
	a_s: add_sub
		generic map
		(
			man_bits			=>	man_bits
		)
		port map
		(
			--inputs
			F1						=>	f1_2,
			F2						=>	f2_2,
			READY					=>	rdy2,
			OP						=>	op_2,
			CLK						=>	CLK,
			--outputs
			F_OUT					=>	f_out,
			OVERFLOW			=>	enable,
			DONE					=>	rdy3
		);

	corr: correction 
		generic map
		(
			exp_bits			=>	exp_bits,
			man_bits			=>	man_bits
		)
		port map
		(
			--inputs
			F							=>	f_out,
			E							=>	e1_3,
			S							=>	s1_3,
			READY					=>	rdy3,
			CLEAR					=>	eq3,
			ENABLE				=>	enable,
			EXCEPTION_IN	=>	exc3,
			CLK						=>	CLK,
			--outputs
			RESULT				=>	RESULT,
			EXCEPTION_OUT	=>	EXCEPTION_OUT,
			DONE					=>	DONE
		);
	
	--permanent connections (signal renaming)
	s1	<=	large(man_bits+exp_bits);
	e1	<=	large(man_bits+exp_bits-1 downto man_bits);
	f1(man_bits downto 1)	<=	large(man_bits-1 downto 0);
	f1(0)									<=	'0';
	s2	<=	small(man_bits+exp_bits);
	e2	<=	small(man_bits+exp_bits-1 downto man_bits);
	f2	<=	small(man_bits-1 downto 0);
	op	<=	s1 XNOR s2;
	n_op2(exp_bits+man_bits-1 downto 0)	<=	OP2(exp_bits+man_bits-1 downto 0);
	n_op2(exp_bits+man_bits)						<=	NOT(OP2(exp_bits+man_bits)); --invert sign
	
	main: process (CLK)
	begin
		if(rising_edge(CLK)) then
			--register creation
			--3rd stage
			s1_3					<=	s1_2;
			e1_3					<=	e1_2;
			exc3					<=	exc2;
			eq3						<=	eq2;
			--2nd stage
			s1_2					<=	s1;
			f1_2					<=	f1;
			op_2					<=	op;
			exc2					<=	exc1;
			eq2						<=	eq1;
			--1st stage
			exc1					<=	EXCEPTION_IN;
			eq1						<=	eq;
		end if;--CLK
	end process;--main
	
end fp_add_arch; --end of architecture

