--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  FILE         | fp_mul.vhd                           --
--  -------------+------------------------------------  --
--  DESCRIPTION  | Floating-point multiplication        --
--               | module.                              --
--  -------------+------------------------------------  --
--  AUTHOR       | Pavle Belanovic                      --
--  -------------+------------------------------------  --
--  DATE         | 20 June 2002                         --
--                                                      --
--======================================================--

--******************************************************************************--
--                                                                              --
--	Copyright (C) 2000 Pavle Belanovic                                          --
--                                                                              --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                              --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                              --
--	You should have received a copy of the GNU General Public License           --
--	along with this program; if not, write to the Free Software                 --
--	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. --
--                                                                              --
--******************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

-- float
library work;
use work.float_pkg.all;

--======================================================--
--                      ENTITIES                        --
--======================================================--

----------------------------------------------------------
--              Floating Point Multiplier               --
----------------------------------------------------------
entity fp_mul is
	generic
	(
		exp_bits			:	integer	:=	0;
		man_bits			:	integer	:=	0
	);
	port
	(
		--inputs
		OP1						:	in		std_logic_vector(exp_bits+man_bits downto 0);
		OP2						:	in		std_logic_vector(exp_bits+man_bits downto 0);
		READY					:	in		std_logic;
		EXCEPTION_IN	:	in		std_logic;
		CLK						:	in		std_logic;
		--outputs
		RESULT				:	out		std_logic_vector(exp_bits+(2*man_bits) downto 0)	:=	(others=>'0');
		EXCEPTION_OUT	:	out		std_logic																					:=	'0';
		DONE					:	out		std_logic																					:=	'0'
	);
end entity;

--======================================================--
--                    ARCHITECTURES                     --
--======================================================--

----------------------------------------------------------
--              Floating Point Multiplier               --
----------------------------------------------------------
architecture fp_mul_arch of fp_mul is
--======================================================--
--                      SIGNALS                         --
--======================================================--
	signal	s1							:	std_logic																					:=	'0';
	signal	e1							:	std_logic_vector(exp_bits-1 downto 0)							:=	(others=>'0');
	signal	f1							:	std_logic_vector(man_bits-1 downto 0)							:=	(others=>'0');
	signal	s2							:	std_logic																					:=	'0';
	signal	e2							:	std_logic_vector(exp_bits-1 downto 0)							:=	(others=>'0');
	signal	f2							:	std_logic_vector(man_bits-1 downto 0)							:=	(others=>'0');
	signal	sign1						:	std_logic																					:=	'0';
	signal	sign2						:	std_logic																					:=	'0';
	signal	rdy2						:	std_logic																					:=	'0';
	signal	rdy3						:	std_logic																					:=	'0';
	signal	exc1						:	std_logic																					:=	'0';
	signal	exc2_a					:	std_logic																					:=	'0';
	signal	exc2_b					:	std_logic																					:=	'0';
	signal	exc2_c					:	std_logic																					:=	'0';
	signal	exc3						:	std_logic																					:=	'0';
	signal	clear						:	std_logic																					:=	'0';
	signal	exp1						: std_logic_vector(exp_bits downto 0)								:=	(others=>'0');
	signal	exp2_a					: std_logic_vector(exp_bits downto 0)								:=	(others=>'0');
	signal	exp2_b					:	std_logic_vector(exp_bits downto 0)								:=	(others=>'0');
	signal	exp2_c					: std_logic_vector(exp_bits-1 downto 0)							:=	(others=>'0');
	signal	bias						:	std_logic_vector(exp_bits downto 0)								:=	(others=>'0');
	signal	man1						:	std_logic_vector((2*man_bits)-1 downto 0)					:=	(others=>'0');
	signal	man2						:	std_logic_vector((2*man_bits)-1 downto 0)					:=	(others=>'0');
	signal	res2						:	std_logic_vector(exp_bits+(2*man_bits) downto 0)	:=	(others=>'0');
	signal	res3						:	std_logic_vector(exp_bits+(2*man_bits) downto 0)	:=	(others=>'0');
	signal	res3_b					:	std_logic_vector(exp_bits+(2*man_bits) downto 0)	:=	(others=>'0');
	signal	zero						:	std_logic_vector(exp_bits+(2*man_bits) downto 0)	:=	(others=>'0');
	signal	op1_or					:	std_logic																					:=	'0';
	signal	op2_or					:	std_logic																					:=	'0';
	signal	all_zero1				:	std_logic																					:=	'0';
	signal	all_zero2				:	std_logic																					:=	'0';
	signal	all_zero3				:	std_logic																					:=	'0';
begin
----------------------------------------------------------
--               Component Instantiation                --
----------------------------------------------------------
	exponent_adder: parameterized_adder
		generic map
		(
			bits	=>	exp_bits
		)
		port map
		(
			A			=>	e1,
			B			=>	e2,
			CIN		=>	'0',
			S			=>	exp1(exp_bits-1 downto 0),
			COUT	=>	exp1(exp_bits)
		);
	mantissa_multiplier: parameterized_multiplier
		generic map
		(
			bits	=>	man_bits
		)
		port map
		(
			A			=>	f1,
			B			=>	f2,
			S			=>	man1
		);
	bias_subtractor: parameterized_subtractor
		generic map
		(
			bits	=>	exp_bits+1
		)
		port map
		(
			A			=>	exp2_a,
			B			=>	bias,
			O			=>	exp2_b
		);
	output_mux: parameterized_mux
		generic map
		(
			bits	=>	1+exp_bits+(2*man_bits)
		)
		port map
		(
			A			=>	zero,
			B			=>	res3,
			S			=>	clear,
			O			=>	res3_b
		);
	all_zero_or_gate_op1: parameterized_or_gate
		generic map
		(
			bits	=>	exp_bits+man_bits
		)
		port map
		(
			A			=>	OP1(exp_bits+man_bits-1 downto 0), --everything but the sign bit
			O			=>	op1_or
		);
	all_zero_or_gate_op2: parameterized_or_gate
		generic map
		(
			bits	=>	exp_bits+man_bits
		)
		port map
		(
			A			=>	OP2(exp_bits+man_bits-1 downto 0), --everything but the sign bit
			O			=>	op2_or
		);
----------------------------------------------------------
--              Asynchronous Assignments                --
----------------------------------------------------------
	s1				<=	OP1(exp_bits+man_bits);
	e1				<=	OP1(exp_bits+man_bits-1 downto man_bits);
	f1				<=	OP1(man_bits-1 downto 0);
	s2				<=	OP2(exp_bits+man_bits);
	e2				<=	OP2(exp_bits+man_bits-1 downto man_bits);
	f2				<=	OP2(man_bits-1 downto 0);
	sign1			<=	s1 XOR s2;
	all_zero1	<=	(NOT(op1_or)) OR (NOT(op2_or));
	exc2_c		<=	exc2_a OR exc2_b;
	exc2_b		<=	exp2_b(exp_bits) AND rdy2 AND (NOT(all_zero2));
	exp2_c		<=	exp2_b(exp_bits-1 downto 0);
	clear			<=	all_zero3 OR exc3 OR (NOT(rdy3));
	res2((2*man_bits)-1 downto 0)											<=	man2;
	res2(exp_bits+(2*man_bits)-1 downto (2*man_bits))	<=	exp2_c;
	res2(exp_bits+(2*man_bits))												<=	sign2;
	bias(exp_bits downto exp_bits-1)									<=	(others=>'0');
	bias(exp_bits-2 downto 1)													<=	(others=>'1');
	bias(0)																						<=	'0';
----------------------------------------------------------
--               Synchronous Assignments                --
----------------------------------------------------------
	synchro: process (CLK)
	begin
		if(rising_edge(CLK)) then
			rdy2					<=	READY;
			exc2_a				<=	EXCEPTION_IN;
			sign2					<=	sign1;
			exp2_a				<=	exp1;
			man2					<=	man1;
			all_zero2			<=	all_zero1;
			rdy3					<=	rdy2;
			exc3					<=	exc2_c;
			res3					<=	res2;
			all_zero3			<=	all_zero2;
			DONE					<=	rdy3;
			EXCEPTION_OUT	<=	exc3;
			RESULT				<=	res3_b;
		end if;--CLK
	end process;--synchro

end fp_mul_arch; --end of architecture

